/* ---------------------------------------------------------------------- */
/*                   Copyright (C) 1991 by Natrlich!                     */
/*                      This file is copyrighted!                         */
/*                Refer to the documentation for details.                 */
/* ---------------------------------------------------------------------- */
#include "defines.h"
#include "nasm.h"
#include OSBIND
#include "code.h"
#include <stddef.h>     /* get size_t */
#include <string.h>
#include <stdlib.h>
#include "debug.h"

#if STATISTICS
#include "stats.h"
#include <stdio.h>
#endif

extern char    outofmem[];

#if STATISTICS || ! VERSION
#define MAXHISTORY   256
#define MAXWATCHES   16
#define MAGIC        0x41424344L

static struct
{
   byte huge   *p;
   lword       size;
} history[ MAXHISTORY];


static struct watch
{
   unsigned long  magic;
   unsigned char  *p;
   char           *name;
   short          size;
   short          occupied;
   unsigned long  chksum;
} watches[ MAXWATCHES] = 
{
   { MAGIC, 0, 0, 0, 0, 0 },   /* fill this up appropriately */
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 },
   { MAGIC, 0, 0, 0, 0, 0 }
};



unsigned long chksum( p, size)
unsigned char   *p;
{
   unsigned long   x;
   int             i;

   x = *p++;
   for( i = 3; i-- && size; size--)
      x = (x << 8) + *p++;
   while( size--)
      x = (x << 3) + *p++;
   return( x);
}


void  add_watch( p, size, name)
unsigned char  *p;
size_t         size;
char           *name;
{
   struct watch  *q;
   int           i;

   ENTER("add_watch");
   if( ! p)
      return;
   for( q = watches, i = MAXWATCHES; i--; q++)
      if( ! q->occupied)
      {
         q->magic    = MAGIC;
         q->p        = p;
         q->name     = name;
         q->size     = size;
         q->chksum   = chksum( p, (int) size);
         q->occupied = 1;
         LEAVE();
         return;
      }
   nierror("Watch space exhausted");
}


print_watches()
{
   struct watch  *q;
   int            i;

   printf("Active watches:\n");
   for( q = watches, i = MAXWATCHES; i--; q++)
      if( q->occupied)
        printf( "\t%s\n", q->name);
}


void del_watch( p)
unsigned char *p;
{
   struct watch  *q;
   int            i;

   ENTER("del_watch");
   if( ! p)
      return;
   for( q = watches, i = MAXWATCHES; i--; q++)
      if( q->occupied && q->p == p)
      {
         q->occupied = 0;
         LEAVE();
         return;
      }
   nierror("Watch not found");
}


do_watch()
{
   struct watch  *q;
   int            i;

   for( q = watches, i = MAXWATCHES; i--; q++)
   {
      if( q->magic != MAGIC)
         nierror("Good work! You managed to clobber the watch space");

      if( q->occupied && q->chksum != chksum( q->p, q->size))
      {
         auto char buffer[ 256];

         if( q->name)
            sprintf( buffer, "Someone clobbered >>%s<<", q->name);
         else
            sprintf( buffer, "Someone clobbered %d bytes @$%.8lX",
                    q->size, (long) q->p);
         nierror( buffer);
      }
   }
}


check_mallocs()
{
   int            i = 0;
   byte huge      *p, huge *q, huge *r;

   for( i = 0; i < MAXHISTORY; i++)
      if( p = (byte huge *) history[ i].p)
      {
         q = p - sizeof( lword);
         r = p + history[ i].size;
         if( lbeek( q) != MAGIC ||
             lbeek( r) != MAGIC)
            nierror("Someone clobbered the malloc space");
      }
}



do_check()
{
   ENTER("do_check");
   do_watch();
   check_mallocs();
   LEAVE();
}



bad_pointer( p)
byte huge   *p;
{
   int         i = 0;
   byte huge   *q;

   if( ! p)
      return;
   for( i = 0; i < MAXHISTORY; i++)
      if( (q = history[ i].p) && p >= q && p < (q + history[ i].size))
         return;
   nierror("You got a bad pointer here");
}

#endif


byte huge  *nmalloc( size)
lword   size;
{
   register byte volatile huge   *p;

   ENTER("nmalloc");
#if ! VERSION
   if( size > 0x01000000)
      nwarning("Probably some conversion trouble when calling nmalloc");
   size += sizeof( lword) * 2;
#endif

#if STATISTICS
   _mused += size;
   if( _mused >= _maxused)
      _maxused = _mused;
#endif
                           /* fucking TURBO.C libraries, nothing works */
#if OS == TOS && __TURBOC__ && ! defined( __PUREC__)
   if( ! (p = (byte huge *) malloc( size)) || (long) p < 0)
#else
   if( ! (p = (byte huge *) malloc( size)))
#endif
      nferror( outofmem);
#if ! VERSION
   memset( p, 0xFF, size);
# if OS == MSDOS
   if( get_ADDR( p) < 0x10000 || get_ADDR( p) > 0x400000)
      nierror( "Pointer trouble in paradise");
# endif
   check_mallocs();
   if( ((*p = 0x55) && *p != 0x55) || ((*p = 0xAA) && *p != 0xAA))
      nierror("malloc is seriously fucked");

   lbyte( p, MAGIC);
   size -= sizeof( lword) * 2;
   p    += sizeof( lword);
   lbyte( p + size, MAGIC);
#endif

#if STATISTICS || ! VERSION
   {
      register int   i = 0;

      while( history[ i].p && i++ < MAXHISTORY);
      if( i == MAXHISTORY)
         nierror("Statistic space for nmalloc exhausted");
      history[ i].p    = (byte huge *) p;
      history[ i].size = size;
   }
#endif
   IMESS("Memory starts @$%8.8lX", (lword) p, 4);
   LEAVE();
   return( (byte huge *) p);
}

void  nfree( p)
void huge   *p;
{
   ENTER("nfree");
   if( ! p)             /* check for p == NULL (Can happen!) */
      return;
#if STATISTICS || ! VERSION
   {
      int   i = 0;

      while( history[ i].p != p && i++ < MAXHISTORY);
      if( i == MAXHISTORY)
         nierror("Faulty nfree");
#if STATISTICS
      _mused -= history[ i].size;
#endif
      memset( (char *) p - sizeof( lword), 0xFE, history[ i].size + 2 * sizeof( lword));
      history[i].p = 0;
   }
#endif
#if ! VERSION
   (char huge *) p -= sizeof( lword);
#endif
   free( p);
   LEAVE();
}


char  *get_filename( pathname)
char  *pathname;
{
   register char  c,
                  *p = pathname;

   while( (c = *p++) && c != DIRSLASH);
   if( ! c)
      return( pathname);
   while( *p++);
   while( *--p != DIRSLASH);
   return( ++p);
}


void  complete( filename, ext, force)
char  *filename, *ext;
{
   char  *x;

   ENTER("complete");

   if( x = strrchr( filename, '.'))
   {
      if( strrchr( filename, DIRSLASH) > x)
         goto dis;
      if( force)
         while( *++x = *++ext);
   }
   else
dis:
      strcat( filename, ext);
   LEAVE();
}

#if LOWERFILE
#include <ctype.h>

void  downcase( s)
register char  *s;
{
   register char  c;
   
   while( c = *s)
      *s++ = tolower( c);
}
#endif
/* ---------------------------------------------------------- */
/* Character representation (from Your Atari Computer)        */
/*   $00-$1F  ->  $40-$5F    (+ $40)                          */
/*   $20-$5F  ->  $00-$3F    (- $20)                          */
/*   $60-$7F  ->  $60-$7F    (+-$00)                          */
/*   $80-$9F  ->  $C0-$DF    (+ $40)                          */
/*   $A0-$DF  ->  $80-$BF    (- $20)                          */
/*   $EO-$FF  ->  $E0-$FF    (+-$00)                          */
/* ---------------------------------------------------------- */
byte  scrtab[] =
{
   0x40, 0x41, 0x42, 0x43,  0x44, 0x45, 0x46, 0x47,
   0x48, 0x49, 0x4A, 0x4B,  0x4C, 0x4D, 0x4E, 0x4F,
   0x50, 0x51, 0x52, 0x53,  0x54, 0x55, 0x56, 0x57,
   0x58, 0x59, 0x5A, 0x5B,  0x5C, 0x5D, 0x5E, 0x5F,
   0x00, 0x01, 0x02, 0x03,  0x04, 0x05, 0x06, 0x07,
   0x08, 0x09, 0x0A, 0x0B,  0x0C, 0x0D, 0x0E, 0x0F,
   0x10, 0x11, 0x12, 0x13,  0x14, 0x15, 0x16, 0x17,
   0x18, 0x19, 0x1A, 0x1B,  0x1C, 0x1D, 0x1E, 0x1F,
   0x20, 0x21, 0x22, 0x23,  0x24, 0x25, 0x26, 0x27,
   0x28, 0x29, 0x2A, 0x2B,  0x2C, 0x2D, 0x2E, 0x2F,
   0x30, 0x31, 0x32, 0x33,  0x34, 0x35, 0x36, 0x37,
   0x38, 0x39, 0x3A, 0x3B,  0x3C, 0x3D, 0x3E, 0x3F,
   0x60, 0x61, 0x62, 0x63,  0x64, 0x65, 0x66, 0x67,
   0x68, 0x69, 0x6A, 0x6B,  0x6C, 0x6D, 0x6E, 0x6F,
   0x70, 0x71, 0x72, 0x73,  0x74, 0x75, 0x76, 0x77,
   0x78, 0x79, 0x7A, 0x7B,  0x7C, 0x7D, 0x7E, 0x7F,

   0xC0, 0xC1, 0xC2, 0xC3,  0xC4, 0xC5, 0xC6, 0xC7,
   0xC8, 0xC9, 0xCA, 0xCB,  0xCC, 0xCD, 0xCE, 0xCF,
   0xD0, 0xD1, 0xD2, 0xD3,  0xD4, 0xD5, 0xD6, 0xD7,
   0xD8, 0xD9, 0xDA, 0xDB,  0xDC, 0xDD, 0xDE, 0xDF,
   0x80, 0x81, 0x82, 0x83,  0x84, 0x85, 0x86, 0x87,
   0x88, 0x89, 0x8A, 0x8B,  0x8C, 0x8D, 0x8E, 0x8F,
   0x90, 0x91, 0x92, 0x93,  0x94, 0x95, 0x96, 0x97,
   0x98, 0x99, 0x9A, 0x9B,  0x9C, 0x9D, 0x9E, 0x9F,
   0xA0, 0xA1, 0xA2, 0xA3,  0xA4, 0xA5, 0xA6, 0xA7,
   0xA8, 0xA9, 0xAA, 0xAB,  0xAC, 0xAD, 0xAE, 0xAF,
   0xB0, 0xB1, 0xB2, 0xB3,  0xB4, 0xB5, 0xB6, 0xB7,
   0xB8, 0xB9, 0xBA, 0xBB,  0xBC, 0xBD, 0xBE, 0xBF,
   0xE0, 0xE1, 0xE2, 0xE3,  0xE4, 0xE5, 0xE6, 0xE7,
   0xE8, 0xE9, 0xEA, 0xEB,  0xEC, 0xED, 0xEE, 0xEF,
   0xF0, 0xF1, 0xF2, 0xF3,  0xF4, 0xF5, 0xF6, 0xF7,
   0xF8, 0xF9, 0xFA, 0xFB,  0xFC, 0xFD, 0xFE, 0xFF
};

